@@ -18,11 +18,15 @@
  * applicable instead of those above.
  */
 
+#include <stdio.h>
+
+#include <iostream>
 #include <map>
 #include <sstream>
 #include <string>
-#include <stdio.h>
-#include <iostream>
+#include <vector>
+
+#include <boost/shared_ptr.hpp>
 
 #include <librevenge-generators/librevenge-generators.h>
 
@@ -31,9 +35,39 @@
 namespace librevenge
 {
 
-namespace
+namespace DrawingSVG
 {
 
+static double getInchValue(librevenge::RVNGProperty const &prop)
+{
+	double value=prop.getDouble();
+	switch (prop.getUnit())
+	{
+	case librevenge::RVNG_GENERIC: // assume inch
+	case librevenge::RVNG_INCH:
+		return value;
+	case librevenge::RVNG_POINT:
+		value /= 72.;
+		return value;
+	case librevenge::RVNG_TWIP:
+		value /= 1440.;
+		return value;
+	case librevenge::RVNG_PERCENT:
+	case librevenge::RVNG_UNIT_ERROR:
+	default:
+	{
+		static bool first=true;
+		if (first)
+		{
+			RVNG_DEBUG_MSG(("librevenge::getInchValue: call with no double value\n"));
+			first=false;
+		}
+		break;
+	}
+	}
+	return value;
+}
+
 static std::string doubleToString(const double value)
 {
 	RVNGProperty *prop = RVNGPropertyFactory::newDoubleProp(value);
@@ -61,7 +95,94 @@ static unsigned stringToColour(const RVNGString &s)
 	return val;
 }
 
-} // anomymous namespace
+//! basic class used to store table information
+struct Table
+{
+	//! constructor
+	Table(const RVNGPropertyList &propList) : m_column(0), m_row(0), m_x(0), m_y(0), m_columnsDistanceFromOrigin(), m_rowsDistanceFromOrigin()
+	{
+		if (propList["svg:x"])
+			m_x=getInchValue(*propList["svg:x"]);
+		if (propList["svg:y"])
+			m_y=getInchValue(*propList["svg:y"]);
+		// we do not actually use height/width, so...
+
+		m_columnsDistanceFromOrigin.push_back(0);
+		m_rowsDistanceFromOrigin.push_back(0);
+
+		const librevenge::RVNGPropertyListVector *columns = propList.child("librevenge:table-columns");
+		if (columns)
+		{
+			double actDist=0;
+			for (unsigned long i=0; i<columns->count(); ++i)
+			{
+				if ((*columns)[i]["style:column-width"])
+					actDist+=getInchValue(*(*columns)[i]["style:column-width"]);
+				m_columnsDistanceFromOrigin.push_back(actDist);
+			}
+		}
+		else
+		{
+			RVNG_DEBUG_MSG(("librevenge::DrawingSVG::Table::Table: can not find any columns\n"));
+		}
+	}
+	//! calls to open a row
+	void openRow(const RVNGPropertyList &propList)
+	{
+		double height=0;
+		if (propList["style:row-height"])
+			height=getInchValue(*propList["style:row-height"]);
+		// changeme
+		else if (propList["style:min-row-height"])
+			height=getInchValue(*propList["style:min-row-height"]);
+		else
+		{
+			RVNG_DEBUG_MSG(("librevenge::DrawingSVG::Table::openRow: can not determine row height\n"));
+		}
+		m_rowsDistanceFromOrigin.push_back(m_rowsDistanceFromOrigin.back()+height);
+	}
+	//! call to close a row
+	void closeRow()
+	{
+		++m_row;
+	}
+	//! returns the position of a cellule
+	bool getPosition(int column, int row, double &x, double &y) const
+	{
+		bool ok=true;
+		if (column>=0 && column <int(m_columnsDistanceFromOrigin.size()))
+			x=m_x+m_columnsDistanceFromOrigin[size_t(column)];
+		else
+		{
+			ok=false;
+			RVNG_DEBUG_MSG(("librevenge::DrawingSVG::Table::getPosition: the column %d seems bad\n", column));
+			x=(column<0 || m_columnsDistanceFromOrigin.empty()) ? m_x : m_x + m_columnsDistanceFromOrigin.back();
+		}
+		if (row>=0 && row <int(m_rowsDistanceFromOrigin.size()))
+			y=m_y+m_rowsDistanceFromOrigin[size_t(row)];
+		else
+		{
+			ok=false;
+			RVNG_DEBUG_MSG(("librevenge::DrawingSVG::Table::getPosition: the row %d seems bad\n", row));
+			y=(row<0 || m_rowsDistanceFromOrigin.empty()) ? m_y : m_y + m_rowsDistanceFromOrigin.back();
+		}
+		return ok;
+	}
+	//! the actual column
+	int m_column;
+	//! the actual row
+	int m_row;
+	//! the origin table position in inches
+	double m_x, m_y;
+	//! the distance of each begin column in inches from origin
+	std::vector<double> m_columnsDistanceFromOrigin;
+	//! the distance of each begin row in inches from origin
+	std::vector<double> m_rowsDistanceFromOrigin;
+};
+
+} // DrawingSVG namespace
+
+using namespace DrawingSVG;
 
 struct RVNGSVGDrawingGeneratorPrivate
 {
@@ -85,6 +206,8 @@ struct RVNGSVGDrawingGeneratorPrivate
 	//! index uses when fill=bitmap
 	int m_patternIndex;
 	int m_arrowStartIndex /** start arrow index*/, m_arrowEndIndex /** end arrow index */;
+	//! groupId used if svg:id is not defined when calling openGroup
+	int m_groupId;
 	//! layerId used if svg:id is not defined when calling startLayer
 	int m_layerId;
 	//! a prefix used to define the svg namespace
@@ -93,6 +216,12 @@ struct RVNGSVGDrawingGeneratorPrivate
 	std::string m_nmSpaceAndDelim;
 	std::ostringstream m_outputSink;
 	RVNGStringVector &m_vec;
+	//! the actual master name
+	RVNGString m_masterName;
+	//! a map master name to master content
+	std::map<RVNGString, std::string> m_masterNameToContentMap;
+	//! the actual opened table
+	boost::shared_ptr<Table> m_table;
 };
 
 RVNGSVGDrawingGeneratorPrivate::RVNGSVGDrawingGeneratorPrivate(RVNGStringVector &vec, const RVNGString &nmSpace) :
@@ -104,11 +233,13 @@ RVNGSVGDrawingGeneratorPrivate::RVNGSVGDrawingGeneratorPrivate(RVNGStringVector
 	m_patternIndex(1),
 	m_arrowStartIndex(1),
 	m_arrowEndIndex(1),
-	m_layerId(1000),
+	m_groupId(1000), m_layerId(1000),
 	m_nmSpace(nmSpace.cstr()),
 	m_nmSpaceAndDelim(""),
 	m_outputSink(),
-	m_vec(vec)
+	m_vec(vec),
+	m_masterName(), m_masterNameToContentMap(),
+	m_table()
 {
 	if (!m_nmSpace.empty())
 		m_nmSpaceAndDelim = m_nmSpace+":";
@@ -124,8 +255,8 @@ void RVNGSVGDrawingGeneratorPrivate::drawPolySomething(const RVNGPropertyListVec
 		if (!vertices[0]["svg:x"]||!vertices[0]["svg:y"]||!vertices[1]["svg:x"]||!vertices[1]["svg:y"])
 			return;
 		m_outputSink << "<" << getNamespaceAndDelim() << "line ";
-		m_outputSink << "x1=\"" << doubleToString(72*(vertices[0]["svg:x"]->getDouble())) << "\"  y1=\"" << doubleToString(72*(vertices[0]["svg:y"]->getDouble())) << "\" ";
-		m_outputSink << "x2=\"" << doubleToString(72*(vertices[1]["svg:x"]->getDouble())) << "\"  y2=\"" << doubleToString(72*(vertices[1]["svg:y"]->getDouble())) << "\"\n";
+		m_outputSink << "x1=\"" << doubleToString(72*getInchValue(*vertices[0]["svg:x"])) << "\"  y1=\"" << doubleToString(72*getInchValue(*vertices[0]["svg:y"])) << "\" ";
+		m_outputSink << "x2=\"" << doubleToString(72*getInchValue(*vertices[1]["svg:x"])) << "\"  y2=\"" << doubleToString(72*getInchValue(*vertices[1]["svg:y"])) << "\"\n";
 		writeStyle();
 		m_outputSink << "/>\n";
 	}
@@ -141,7 +272,7 @@ void RVNGSVGDrawingGeneratorPrivate::drawPolySomething(const RVNGPropertyListVec
 		{
 			if (!vertices[i]["svg:x"]||!vertices[i]["svg:y"])
 				continue;
-			m_outputSink << doubleToString(72*(vertices[i]["svg:x"]->getDouble())) << " " << doubleToString(72*(vertices[i]["svg:y"]->getDouble()));
+			m_outputSink << doubleToString(72*getInchValue(*vertices[i]["svg:x"])) << " " << doubleToString(72*getInchValue(*vertices[i]["svg:y"]));
 			if (i < vertices.count()-1)
 				m_outputSink << ", ";
 		}
@@ -176,9 +307,9 @@ void RVNGSVGDrawingGeneratorPrivate::setStyle(const RVNGPropertyList &propList)
 		m_outputSink << "<" << getNamespaceAndDelim() << "filter filterUnits=\"userSpaceOnUse\" id=\"shadow" << m_shadowIndex++ << "\">";
 		m_outputSink << "<" << getNamespaceAndDelim() << "feOffset in=\"SourceGraphic\" result=\"offset\" ";
 		if (m_style["draw:shadow-offset-x"])
-			m_outputSink << "dx=\"" << doubleToString(72*m_style["draw:shadow-offset-x"]->getDouble()) << "\" ";
+			m_outputSink << "dx=\"" << doubleToString(72*getInchValue(*m_style["draw:shadow-offset-x"])) << "\" ";
 		if (m_style["draw:shadow-offset-y"])
-			m_outputSink << "dy=\"" << doubleToString(72*m_style["draw:shadow-offset-y"]->getDouble()) << "\" ";
+			m_outputSink << "dy=\"" << doubleToString(72*getInchValue(*m_style["draw:shadow-offset-y"])) << "\" ";
 		m_outputSink << "/>";
 		m_outputSink << "<" << getNamespaceAndDelim() << "feColorMatrix in=\"offset\" result=\"offset-color\" type=\"matrix\" values=\"";
 		m_outputSink << "0 0 0 0 " << doubleToString(shadowRed) ;
@@ -378,33 +509,33 @@ void RVNGSVGDrawingGeneratorPrivate::setStyle(const RVNGPropertyList &propList)
 		m_outputSink << "<" << getNamespaceAndDelim() << "defs>\n";
 		m_outputSink << "  <" << getNamespaceAndDelim() << "pattern id=\"img" << m_patternIndex++ << "\" patternUnits=\"userSpaceOnUse\" ";
 		if (m_style["svg:width"])
-			m_outputSink << "width=\"" << doubleToString(72*(m_style["svg:width"]->getDouble())) << "\" ";
+			m_outputSink << "width=\"" << doubleToString(72*getInchValue(*m_style["svg:width"])) << "\" ";
 		else
 			m_outputSink << "width=\"100\" ";
 
 		if (m_style["svg:height"])
-			m_outputSink << "height=\"" << doubleToString(72*(m_style["svg:height"]->getDouble())) << "\">" << std::endl;
+			m_outputSink << "height=\"" << doubleToString(72*getInchValue(*m_style["svg:height"])) << "\">" << std::endl;
 		else
 			m_outputSink << "height=\"100\">" << std::endl;
 		m_outputSink << "<" << getNamespaceAndDelim() << "image ";
 
 		if (m_style["svg:x"])
-			m_outputSink << "x=\"" << doubleToString(72*(m_style["svg:x"]->getDouble())) << "\" ";
+			m_outputSink << "x=\"" << doubleToString(72*getInchValue(*m_style["svg:x"])) << "\" ";
 		else
 			m_outputSink << "x=\"0\" ";
 
 		if (m_style["svg:y"])
-			m_outputSink << "y=\"" << doubleToString(72*(m_style["svg:y"]->getDouble())) << "\" ";
+			m_outputSink << "y=\"" << doubleToString(72*getInchValue(*m_style["svg:y"])) << "\" ";
 		else
 			m_outputSink << "y=\"0\" ";
 
 		if (m_style["svg:width"])
-			m_outputSink << "width=\"" << doubleToString(72*(m_style["svg:width"]->getDouble())) << "\" ";
+			m_outputSink << "width=\"" << doubleToString(72*getInchValue(*m_style["svg:width"])) << "\" ";
 		else
 			m_outputSink << "width=\"100\" ";
 
 		if (m_style["svg:height"])
-			m_outputSink << "height=\"" << doubleToString(72*(m_style["svg:height"]->getDouble())) << "\" ";
+			m_outputSink << "height=\"" << doubleToString(72*getInchValue(*m_style["svg:height"])) << "\" ";
 		else
 			m_outputSink << "height=\"100\" ";
 
@@ -446,7 +577,7 @@ void RVNGSVGDrawingGeneratorPrivate::writeStyle(bool /* isClosed */)
 	double width = 1.0 / 72.0;
 	if (m_style["svg:stroke-width"])
 	{
-		width = m_style["svg:stroke-width"]->getDouble();
+		width = getInchValue(*m_style["svg:stroke-width"]);
 #if 0
 		// add me in libmspub and libcdr
 		if (width <= 0.0 && m_style["draw:stroke"] && m_style["draw:stroke"]->getStr() != "none")
@@ -472,24 +603,24 @@ void RVNGSVGDrawingGeneratorPrivate::writeStyle(bool /* isClosed */)
 		double dots1len = 72.*width, dots2len = 72.*width, gap = 72.*width;
 		if (m_style["draw:dots1-length"])
 		{
-			dots1len = 72.*m_style["draw:dots1-length"]->getDouble();
-			std::string str = m_style["draw:dots1-length"]->getStr().cstr();
-			if (str.size() > 1 && str[str.size()-1]=='%')
-				dots1len *=width;
+			if (m_style["draw:dots1-length"]->getUnit()==librevenge::RVNG_PERCENT)
+				dots1len=72*m_style["draw:dots1-length"]->getDouble()*width;
+			else
+				dots1len=72*getInchValue(*m_style["draw:dots1-length"]);
 		}
 		if (m_style["draw:dots2-length"])
 		{
-			dots2len = 72.*m_style["draw:dots2-length"]->getDouble();
-			std::string str = m_style["draw:dots2-length"]->getStr().cstr();
-			if (str.size() > 1 && str[str.size()-1]=='%')
-				dots2len *=width;
+			if (m_style["draw:dots2-length"]->getUnit()==librevenge::RVNG_PERCENT)
+				dots2len=72*m_style["draw:dots2-length"]->getDouble()*width;
+			else
+				dots2len=72*getInchValue(*m_style["draw:dots2-length"]);
 		}
 		if (m_style["draw:distance"])
 		{
-			gap = 72.*m_style["draw:distance"]->getDouble();
-			std::string str = m_style["draw:distance"]->getStr().cstr();
-			if (str.size() > 1 && str[str.size()-1]=='%')
-				gap *=width;
+			if (m_style["draw:distance"]->getUnit()==librevenge::RVNG_PERCENT)
+				gap=72*m_style["draw:distance"]->getDouble()*width;
+			else
+				gap=72*getInchValue(*m_style["draw:distance"]);
 		}
 		m_outputSink << "stroke-dasharray: ";
 		for (int i = 0; i < dots1; i++)
@@ -561,18 +692,24 @@ void RVNGSVGDrawingGenerator::defineEmbeddedFont(const RVNGPropertyList & /*prop
 
 void RVNGSVGDrawingGenerator::startPage(const RVNGPropertyList &propList)
 {
-#if 0
-	m_pImpl->m_outputSink << "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n";
-	m_pImpl->m_outputSink << "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\"";
-	m_pImpl->m_outputSink << " \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n";
-#endif
+	if (propList["librevenge:master-page-name"])
+	{
+		if (m_pImpl->m_masterNameToContentMap.find(propList["librevenge:master-page-name"]->getStr())!=
+		        m_pImpl->m_masterNameToContentMap.end())
+		{
+			m_pImpl->m_outputSink << m_pImpl->m_masterNameToContentMap.find(propList["librevenge:master-page-name"]->getStr())->second;
+			return;
+		}
+		RVNG_DEBUG_MSG(("RVNGSVGDrawingGenerator::startPage: can not find page with given master name\n"));
+	}
+
 	m_pImpl->m_outputSink << "<" << m_pImpl->getNamespaceAndDelim() << "svg version=\"1.1\" xmlns";
 	m_pImpl->m_outputSink << (m_pImpl->m_nmSpace.empty() ? "" : ":") << m_pImpl->m_nmSpace << "=\"http://www.w3.org/2000/svg\" ";
 	m_pImpl->m_outputSink << "xmlns:xlink=\"http://www.w3.org/1999/xlink\" ";
 	if (propList["svg:width"])
-		m_pImpl->m_outputSink << "width=\"" << doubleToString(72*(propList["svg:width"]->getDouble())) << "\" ";
+		m_pImpl->m_outputSink << "width=\"" << doubleToString(72*getInchValue(*propList["svg:width"])) << "\" ";
 	if (propList["svg:height"])
-		m_pImpl->m_outputSink << "height=\"" << doubleToString(72*(propList["svg:height"]->getDouble())) << "\"";
+		m_pImpl->m_outputSink << "height=\"" << doubleToString(72*getInchValue(*propList["svg:height"])) << "\"";
 	m_pImpl->m_outputSink << " >\n";
 }
 
@@ -583,23 +720,59 @@ void RVNGSVGDrawingGenerator::endPage()
 	m_pImpl->m_outputSink.str("");
 }
 
-void RVNGSVGDrawingGenerator::startMasterPage(const RVNGPropertyList &)
+void RVNGSVGDrawingGenerator::startMasterPage(const RVNGPropertyList &propList)
 {
+	if (!m_pImpl->m_masterName.empty())
+	{
+		RVNG_DEBUG_MSG(("RVNGSVGDrawingGenerator::startMasterPage: a master page is already opened\n"));
+	}
+	else if (propList["librevenge:master-page-name"])
+	{
+		m_pImpl->m_masterName=propList["librevenge:master-page-name"]->getStr();
+		RVNGPropertyList pList(propList);
+		pList.remove("librevenge:master-page-name");
+		startPage(pList);
+	}
+	else
+	{
+		RVNG_DEBUG_MSG(("RVNGSVGDrawingGenerator::startMasterPage: can not find the master name\n"));
+	}
 }
 
 void RVNGSVGDrawingGenerator::endMasterPage()
 {
-	// we don't do anything with master pages yet, so just reset the content
+	if (m_pImpl->m_masterName.empty())
+	{
+		RVNG_DEBUG_MSG(("RVNGSVGDrawingGenerator::endMasterPage: no master pages are opened\n"));
+	}
+	else
+	{
+		if (m_pImpl->m_masterNameToContentMap.find(m_pImpl->m_masterName)!=m_pImpl->m_masterNameToContentMap.end())
+		{
+			RVNG_DEBUG_MSG(("RVNGSVGDrawingGenerator::endMasterPage: a master page with name %s already exists\n",
+			                m_pImpl->m_masterName.cstr()));
+		}
+		// no need to close the page, this will be done when the master page will be used
+		m_pImpl->m_masterNameToContentMap[m_pImpl->m_masterName]=m_pImpl->m_outputSink.str();
+		m_pImpl->m_masterName.clear();
+	}
+	// reset the content
 	m_pImpl->m_outputSink.str("");
 }
 
 void RVNGSVGDrawingGenerator::startLayer(const RVNGPropertyList &propList)
 {
 	m_pImpl->m_outputSink << "<" << m_pImpl->getNamespaceAndDelim() << "g";
-	if (propList["svg:id"])
-		m_pImpl->m_outputSink << " id=\"Layer" << propList["svg:id"]->getStr().cstr() << "\"";
+	librevenge::RVNGString layer("Layer");
+	if (propList["draw:layer"])
+		layer.append(propList["draw:layer"]->getStr());
+	else if (propList["svg:id"])
+		layer.append(propList["svg:id"]->getStr());
 	else
-		m_pImpl->m_outputSink << " id=\"Layer" << m_pImpl->m_layerId++ << "\"";
+		layer.sprintf("Layer%d", m_pImpl->m_layerId++);
+	librevenge::RVNGString finalName("");
+	finalName.appendEscapedXML(layer);
+	m_pImpl->m_outputSink << " id=\"" << finalName.cstr() << "\"";
 	if (propList["svg:fill-rule"])
 		m_pImpl->m_outputSink << " fill-rule=\"" << propList["svg:fill-rule"]->getStr().cstr() << "\"";
 	m_pImpl->m_outputSink << " >\n";
@@ -613,8 +786,19 @@ void RVNGSVGDrawingGenerator::endLayer()
 void RVNGSVGDrawingGenerator::startEmbeddedGraphics(const RVNGPropertyList & /*propList*/) {}
 void RVNGSVGDrawingGenerator::endEmbeddedGraphics() {}
 
-void RVNGSVGDrawingGenerator::openGroup(const RVNGPropertyList & /*propList*/) {}
-void RVNGSVGDrawingGenerator::closeGroup() {}
+void RVNGSVGDrawingGenerator::openGroup(const RVNGPropertyList & /*propList*/)
+{
+	m_pImpl->m_outputSink << "<" << m_pImpl->getNamespaceAndDelim() << "g";
+	librevenge::RVNGString group;
+	group.sprintf("Group%d", m_pImpl->m_groupId++);
+	m_pImpl->m_outputSink << " id=\"" << group.cstr() << "\"";
+	m_pImpl->m_outputSink << " >\n";
+}
+
+void RVNGSVGDrawingGenerator::closeGroup()
+{
+	m_pImpl->m_outputSink << "</" << m_pImpl->getNamespaceAndDelim() << "g>\n";
+}
 
 void RVNGSVGDrawingGenerator::setStyle(const RVNGPropertyList &propList)
 {
@@ -626,10 +810,10 @@ void RVNGSVGDrawingGenerator::drawRectangle(const RVNGPropertyList &propList)
 	if (!propList["svg:x"] || !propList["svg:y"] || !propList["svg:width"] || !propList["svg:height"])
 		return;
 	m_pImpl->m_outputSink << "<" << m_pImpl->getNamespaceAndDelim() << "rect ";
-	m_pImpl->m_outputSink << "x=\"" << doubleToString(72*propList["svg:x"]->getDouble()) << "\" y=\"" << doubleToString(72*propList["svg:y"]->getDouble()) << "\" ";
-	m_pImpl->m_outputSink << "width=\"" << doubleToString(72*propList["svg:width"]->getDouble()) << "\" height=\"" << doubleToString(72*propList["svg:height"]->getDouble()) << "\" ";
+	m_pImpl->m_outputSink << "x=\"" << doubleToString(72*getInchValue(*propList["svg:x"])) << "\" y=\"" << doubleToString(72*getInchValue(*propList["svg:y"])) << "\" ";
+	m_pImpl->m_outputSink << "width=\"" << doubleToString(72*getInchValue(*propList["svg:width"])) << "\" height=\"" << doubleToString(72*getInchValue(*propList["svg:height"])) << "\" ";
 	if (propList["svg:rx"] && propList["svg:rx"]->getDouble() > 0 && propList["svg:ry"] && propList["svg:ry"]->getDouble()>0)
-		m_pImpl->m_outputSink << "rx=\"" << doubleToString(72*propList["svg:rx"]->getDouble()) << "\" ry=\"" << doubleToString(72*propList["svg:ry"]->getDouble()) << "\" ";
+		m_pImpl->m_outputSink << "rx=\"" << doubleToString(72*getInchValue(*propList["svg:rx"])) << "\" ry=\"" << doubleToString(72*getInchValue(*propList["svg:ry"])) << "\" ";
 	m_pImpl->writeStyle();
 	m_pImpl->m_outputSink << "/>\n";
 }
@@ -639,13 +823,13 @@ void RVNGSVGDrawingGenerator::drawEllipse(const RVNGPropertyList &propList)
 	if (!propList["svg:cx"] || !propList["svg:cy"] || !propList["svg:rx"] || !propList["svg:ry"])
 		return;
 	m_pImpl->m_outputSink << "<" << m_pImpl->getNamespaceAndDelim() << "ellipse ";
-	m_pImpl->m_outputSink << "cx=\"" << doubleToString(72*propList["svg:cx"]->getDouble()) << "\" cy=\"" << doubleToString(72*propList["svg:cy"]->getDouble()) << "\" ";
-	m_pImpl->m_outputSink << "rx=\"" << doubleToString(72*propList["svg:rx"]->getDouble()) << "\" ry=\"" << doubleToString(72*propList["svg:ry"]->getDouble()) << "\" ";
+	m_pImpl->m_outputSink << "cx=\"" << doubleToString(72*getInchValue(*propList["svg:cx"])) << "\" cy=\"" << doubleToString(72*getInchValue(*propList["svg:cy"])) << "\" ";
+	m_pImpl->m_outputSink << "rx=\"" << doubleToString(72*getInchValue(*propList["svg:rx"])) << "\" ry=\"" << doubleToString(72*getInchValue(*propList["svg:ry"])) << "\" ";
 	m_pImpl->writeStyle();
 	if (propList["librevenge:rotate"] && (propList["librevenge:rotate"]->getDouble()<0||propList["librevenge:rotate"]->getDouble()>0))
 		m_pImpl->m_outputSink << " transform=\" rotate(" << doubleToString(-propList["librevenge:rotate"]->getDouble())
-		                      << ", " << doubleToString(72*propList["svg:cy"]->getDouble())
-		                      << ", " << doubleToString(72*propList["svg:cy"]->getDouble())
+		                      << ", " << doubleToString(72*getInchValue(*propList["svg:cx"]))
+		                      << ", " << doubleToString(72*getInchValue(*propList["svg:cy"]))
 		                      << ")\" ";
 	m_pImpl->m_outputSink << "/>\n";
 }
@@ -682,35 +866,35 @@ void RVNGSVGDrawingGenerator::drawPath(const RVNGPropertyList &propList)
 		bool coord1Ok=coordOk && pList["svg:x1"]&&pList["svg:y1"];
 		bool coord2Ok=coord1Ok && pList["svg:x2"]&&pList["svg:y2"];
 		if (pList["svg:x"] && action[0] == 'H')
-			m_pImpl->m_outputSink << "\nH" << doubleToString(72*(pList["svg:x"]->getDouble()));
+			m_pImpl->m_outputSink << "\nH" << doubleToString(72*getInchValue(*pList["svg:x"]));
 		else if (pList["svg:y"] && action[0] == 'V')
-			m_pImpl->m_outputSink << "\nV" << doubleToString(72*(pList["svg:y"]->getDouble()));
+			m_pImpl->m_outputSink << "\nV" << doubleToString(72*getInchValue(*pList["svg:y"]));
 		else if (coordOk && (action[0] == 'M' || action[0] == 'L' || action[0] == 'T'))
 		{
 			m_pImpl->m_outputSink << "\n" << action;
-			m_pImpl->m_outputSink << doubleToString(72*(pList["svg:x"]->getDouble())) << "," << doubleToString(72*(pList["svg:y"]->getDouble()));
+			m_pImpl->m_outputSink << doubleToString(72*getInchValue(*pList["svg:x"])) << "," << doubleToString(72*getInchValue(*pList["svg:y"]));
 		}
 		else if (coord1Ok && (action[0] == 'Q' || action[0] == 'S'))
 		{
 			m_pImpl->m_outputSink << "\n" << action;
-			m_pImpl->m_outputSink << doubleToString(72*(pList["svg:x1"]->getDouble())) << "," << doubleToString(72*(pList["svg:y1"]->getDouble())) << " ";
-			m_pImpl->m_outputSink << doubleToString(72*(pList["svg:x"]->getDouble())) << "," << doubleToString(72*(pList["svg:y"]->getDouble()));
+			m_pImpl->m_outputSink << doubleToString(72*getInchValue(*pList["svg:x1"])) << "," << doubleToString(72*getInchValue(*pList["svg:y1"])) << " ";
+			m_pImpl->m_outputSink << doubleToString(72*getInchValue(*pList["svg:x"])) << "," << doubleToString(72*getInchValue(*pList["svg:y"]));
 		}
 		else if (coord2Ok && action[0] == 'C')
 		{
 			m_pImpl->m_outputSink << "\nC";
-			m_pImpl->m_outputSink << doubleToString(72*(pList["svg:x1"]->getDouble())) << "," << doubleToString(72*(pList["svg:y1"]->getDouble())) << " ";
-			m_pImpl->m_outputSink << doubleToString(72*(pList["svg:x2"]->getDouble())) << "," << doubleToString(72*(pList["svg:y2"]->getDouble())) << " ";
-			m_pImpl->m_outputSink << doubleToString(72*(pList["svg:x"]->getDouble())) << "," << doubleToString(72*(pList["svg:y"]->getDouble()));
+			m_pImpl->m_outputSink << doubleToString(72*getInchValue(*pList["svg:x1"])) << "," << doubleToString(72*getInchValue(*pList["svg:y1"])) << " ";
+			m_pImpl->m_outputSink << doubleToString(72*getInchValue(*pList["svg:x2"])) << "," << doubleToString(72*getInchValue(*pList["svg:y2"])) << " ";
+			m_pImpl->m_outputSink << doubleToString(72*getInchValue(*pList["svg:x"])) << "," << doubleToString(72*getInchValue(*pList["svg:y"]));
 		}
 		else if (coordOk && pList["svg:rx"] && pList["svg:ry"] && action[0] == 'A')
 		{
 			m_pImpl->m_outputSink << "\nA";
-			m_pImpl->m_outputSink << doubleToString(72*(pList["svg:rx"]->getDouble())) << "," << doubleToString(72*(pList["svg:ry"]->getDouble())) << " ";
+			m_pImpl->m_outputSink << doubleToString(72*getInchValue(*pList["svg:rx"])) << "," << doubleToString(72*getInchValue(*pList["svg:ry"])) << " ";
 			m_pImpl->m_outputSink << doubleToString(pList["librevenge:rotate"] ? pList["librevenge:rotate"]->getDouble() : 0) << " ";
 			m_pImpl->m_outputSink << (pList["librevenge:large-arc"] ? pList["librevenge:large-arc"]->getInt() : 1) << ",";
 			m_pImpl->m_outputSink << (pList["librevenge:sweep"] ? pList["librevenge:sweep"]->getInt() : 1) << " ";
-			m_pImpl->m_outputSink << doubleToString(72*(pList["svg:x"]->getDouble())) << "," << doubleToString(72*(pList["svg:y"]->getDouble()));
+			m_pImpl->m_outputSink << doubleToString(72*getInchValue(*pList["svg:x"])) << "," << doubleToString(72*getInchValue(*pList["svg:y"]));
 		}
 		else if (action[0] == 'Z')
 		{
@@ -733,10 +917,10 @@ void RVNGSVGDrawingGenerator::drawGraphicObject(const RVNGPropertyList &propList
 	m_pImpl->m_outputSink << "<" << m_pImpl->getNamespaceAndDelim() << "image ";
 	if (propList["svg:x"] && propList["svg:y"] && propList["svg:width"] && propList["svg:height"])
 	{
-		double x(propList["svg:x"]->getDouble());
-		double y(propList["svg:y"]->getDouble());
-		double width(propList["svg:width"]->getDouble());
-		double height(propList["svg:height"]->getDouble());
+		double x=getInchValue(*propList["svg:x"]);
+		double y=getInchValue(*propList["svg:y"]);
+		double width=getInchValue(*propList["svg:width"]);
+		double height=getInchValue(*propList["svg:height"]);
 		bool flipX(propList["draw:mirror-horizontal"] && propList["draw:mirror-horizontal"]->getInt());
 		bool flipY(propList["draw:mirror-vertical"] && propList["draw:mirror-vertical"]->getInt());
 
@@ -781,8 +965,8 @@ void RVNGSVGDrawingGenerator::startTextObject(const RVNGPropertyList &propList)
 	m_pImpl->m_outputSink << "<" << m_pImpl->getNamespaceAndDelim() << "text ";
 	if (propList["svg:x"] && propList["svg:y"])
 	{
-		x = propList["svg:x"]->getDouble();
-		y = propList["svg:y"]->getDouble();
+		x = getInchValue(*propList["svg:x"]);
+		y = getInchValue(*propList["svg:y"]);
 	}
 
 	double xmiddle = x;
@@ -790,13 +974,13 @@ void RVNGSVGDrawingGenerator::startTextObject(const RVNGPropertyList &propList)
 
 	if (propList["svg:width"])
 	{
-		double width = propList["svg:width"]->getDouble();
+		double width = getInchValue(*propList["svg:width"]);
 		xmiddle += width / 2.0;
 	}
 
 	if (propList["svg:height"])
 	{
-		height = propList["svg:height"]->getDouble();
+		height = getInchValue(*propList["svg:height"]);
 		ymiddle += height / 2.0;
 	}
 
@@ -920,39 +1104,70 @@ void RVNGSVGDrawingGenerator::insertLineBreak()
 
 void RVNGSVGDrawingGenerator::insertField(const RVNGPropertyList & /*propList*/) {}
 
-void RVNGSVGDrawingGenerator::startTableObject(const RVNGPropertyList &/*propList*/)
+void RVNGSVGDrawingGenerator::startTableObject(const RVNGPropertyList &propList)
 {
-	// TODO: implement me
+	if (m_pImpl->m_table)
+	{
+		RVNG_DEBUG_MSG(("RVNGSVGDrawingGenerator::startTableObject: a table is already opened\n"));
+		return;
+	}
+	m_pImpl->m_table.reset(new Table(propList));
 }
 
-void RVNGSVGDrawingGenerator::openTableRow(const RVNGPropertyList &/*propList*/)
+void RVNGSVGDrawingGenerator::openTableRow(const RVNGPropertyList &propList)
 {
-	// TODO: implement me
+	if (!m_pImpl->m_table) return;
+	m_pImpl->m_table->openRow(propList);
 }
 
 void RVNGSVGDrawingGenerator::closeTableRow()
 {
-	// TODO: implement me
+	if (!m_pImpl->m_table) return;
+	m_pImpl->m_table->closeRow();
 }
 
-void RVNGSVGDrawingGenerator::openTableCell(const RVNGPropertyList &/*propList*/)
+void RVNGSVGDrawingGenerator::openTableCell(const RVNGPropertyList &propList)
 {
-	// TODO: implement me
+	if (!m_pImpl->m_table) return;
+
+	if (propList["librevenge:column"])
+		m_pImpl->m_table->m_column=propList["librevenge:column"]->getInt();
+	if (propList["librevenge:row"])
+		m_pImpl->m_table->m_row=propList["librevenge:row"]->getInt();
+
+	double x = 0, y=0;
+	m_pImpl->m_table->getPosition(m_pImpl->m_table->m_column, m_pImpl->m_table->m_row, x, y);
+	m_pImpl->m_outputSink << "<" << m_pImpl->getNamespaceAndDelim() << "text ";
+	m_pImpl->m_outputSink << "x=\"" << doubleToString(72*x) << "\" y=\"" << doubleToString(72*y) << "\"";
+	m_pImpl->m_outputSink << ">\n";
+
+	// time to update the next cell's column
+	if (propList["table:number-columns-spanned"])
+		m_pImpl->m_table->m_column += propList["librevenge:column"]->getInt();
+	else
+		++m_pImpl->m_table->m_column;
 }
 
 void RVNGSVGDrawingGenerator::closeTableCell()
 {
-	// TODO: implement me
+	if (!m_pImpl->m_table) return;
+	m_pImpl->m_outputSink << "</" << m_pImpl->getNamespaceAndDelim() << "text>\n";
 }
 
 void RVNGSVGDrawingGenerator::insertCoveredTableCell(const RVNGPropertyList &/*propList*/)
 {
+	if (!m_pImpl->m_table) return;
 	// TODO: implement me
 }
 
 void RVNGSVGDrawingGenerator::endTableObject()
 {
-	// TODO: implement me
+	if (!m_pImpl->m_table)
+	{
+		RVNG_DEBUG_MSG(("RVNGSVGDrawingGenerator::endTableObject: no table is already opened\n"));
+		return;
+	}
+	m_pImpl->m_table.reset();
 }
 
 }
